#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Dart interface
# Paul Dechamps

import numpy as np

import dart.utils as floU
import tbox.utils as tboxU
from ..viscous.DAirfoil import Airfoil
from ..viscous.DWake import Wake
from ..viscous import DGroupBuilder

class DartInterface:
    def __init__(self, dartSolver, vSolver, blw):
        self.solver = dartSolver
        self.vSolver = vSolver
        self.createProblem(blw[0], blw[1])
        self.nElmUpperPrev = 0

    def createProblem(self, _boundaryAirfoil, _boundaryWake):
        """Create data structure for information transfer.
        
        Args
        ----
        - _boundaryAirfoil: dart::Blowing data structure for the airfoil.
        - _boundaryWake: dart::Blowing data structure for the wake.
        """
        self.group = [Airfoil(_boundaryAirfoil), Wake(_boundaryWake)]
        self.blwVel = [None, None, None]
        self.deltaStar = [None, None, None]
        self.xx = [None, None, None]
    
    def run(self):
        return self.solver.run()
    
    def writeCp(self, sfx=''):
        """ Write Cp distribution around the airfoil on file.
        """
        Cp = floU.extract(self.group[0].boundary.tag.elems, self.solver.Cp)
        tboxU.write(Cp, 'Cp'+sfx+'.dat', '%1.5e',', ', 'x, y, z, Cp', '')
    
    def save(self, writer):
        self.solver.save(writer)

    def getAoA(self):
        return self.solver.pbl.alpha
    
    def getMinf(self):
        return self.solver.pbl.M_inf
    
    def getCl(self):
        return self.solver.Cl
    
    def getCd(self):
        return self.solver.Cd

    def getCm(self):
        return self.solver.Cm
    
    def getVerbose(self):
        return self.solver.verbose
    
    def reset(self):
        self.solver.reset()
    
    def imposeInvBC(self, couplIter):
        """ Extract the inviscid paramaters at the edge of the boundary layer
        and use it as a boundary condition in the viscous solver.

        Params
        ------
        - couplIter (int): Coupling iteration.
        """
        # Retreive parameters.
        for n in range(0, len(self.group)):
            for i in range(0, len(self.group[n].boundary.nodes)):
                self.group[n].v[i,0] = self.solver.U[self.group[n].boundary.nodes[i].row][0]
                self.group[n].v[i,1] = self.solver.U[self.group[n].boundary.nodes[i].row][1]
                self.group[n].v[i,2] = self.solver.U[self.group[n].boundary.nodes[i].row][2]
                self.group[n].Me[i] = self.solver.M[self.group[n].boundary.nodes[i].row]
                self.group[n].rhoe[i] = self.solver.rho[self.group[n].boundary.nodes[i].row]

        dataIn = [None, None]
        for n in range(0, len(self.group)):
            self.group[n].connectListNodes, self.group[n].connectListElems, dataIn[n] = self.group[n].connectList(couplIter)
        self.fMeshDict, _, _ = DGroupBuilder.mergeVectors(dataIn, 0, 1)
        fMeshDict = self.fMeshDict.copy()

        if ((len(fMeshDict['locCoord'][:fMeshDict['bNodes'][1]]) != self.nElmUpperPrev) or couplIter == 0):
            # Initialize mesh
            self.vSolver.setGlobMesh(0, 0, fMeshDict['globCoord'][:fMeshDict['bNodes'][1]], fMeshDict['yGlobCoord'][:fMeshDict['bNodes'][1]], fMeshDict['zGlobCoord'][:fMeshDict['bNodes'][1]])
            self.vSolver.setGlobMesh(0, 1, fMeshDict['globCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['yGlobCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['zGlobCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
            self.vSolver.setGlobMesh(0, 2, fMeshDict['globCoord'][fMeshDict['bNodes'][2]:], fMeshDict['yGlobCoord'][fMeshDict['bNodes'][2]:], fMeshDict['zGlobCoord'][fMeshDict['bNodes'][2]:])

            # Initialize parameters at the edge of the boundary layer.
            self.vSolver.setxxExt(0, 0, fMeshDict['locCoord'][:fMeshDict['bNodes'][1]])
            self.vSolver.setxxExt(0, 1, fMeshDict['locCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
            self.vSolver.setxxExt(0, 2, fMeshDict['locCoord'][fMeshDict['bNodes'][2]:])
            self.vSolver.setDeltaStarExt(0, 0, fMeshDict['deltaStarExt'][:fMeshDict['bNodes'][1]])
            self.vSolver.setDeltaStarExt(0, 1, fMeshDict['deltaStarExt'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
            self.vSolver.setDeltaStarExt(0, 2, fMeshDict['deltaStarExt'][fMeshDict['bNodes'][2]:])
        else:
            # Parameters at the edge of the boundary layer only.
            self.vSolver.setxxExt(0, 0, fMeshDict['xxExt'][:fMeshDict['bNodes'][1]])
            self.vSolver.setxxExt(0, 1, fMeshDict['xxExt'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
            self.vSolver.setxxExt(0, 2, fMeshDict['xxExt'][fMeshDict['bNodes'][2]:])
            self.vSolver.setDeltaStarExt(0, 0, fMeshDict['deltaStarExt'][:fMeshDict['bNodes'][1]])
            self.vSolver.setDeltaStarExt(0, 1, fMeshDict['deltaStarExt'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
            self.vSolver.setDeltaStarExt(0, 2, fMeshDict['deltaStarExt'][fMeshDict['bNodes'][2]:])

        # Inviscid boundary condition.
        self.nElmUpperPrev = len(fMeshDict['locCoord'][:fMeshDict['bNodes'][1]])
        self.vSolver.setVelocities(0, 0, fMeshDict['vx'][:fMeshDict['bNodes'][1]], fMeshDict['vy'][:fMeshDict['bNodes'][1]], fMeshDict['vz'][:fMeshDict['bNodes'][1]])
        self.vSolver.setVelocities(0, 1, fMeshDict['vx'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['vy'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['vz'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
        self.vSolver.setVelocities(0, 2, fMeshDict['vx'][fMeshDict['bNodes'][2]:], fMeshDict['vy'][fMeshDict['bNodes'][2]:], fMeshDict['vz'][fMeshDict['bNodes'][2]:])
        self.vSolver.setMe(0, 0, fMeshDict['Me'][:fMeshDict['bNodes'][1]])
        self.vSolver.setMe(0, 1, fMeshDict['Me'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
        self.vSolver.setMe(0, 2, fMeshDict['Me'][fMeshDict['bNodes'][2]:])
        self.vSolver.setRhoe(0, 0, fMeshDict['rho'][:fMeshDict['bNodes'][1]])
        self.vSolver.setRhoe(0, 1, fMeshDict['rho'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
        self.vSolver.setRhoe(0, 2, fMeshDict['rho'][fMeshDict['bNodes'][2]:])

    def imposeBlwVel(self):
        """ Extract the solution of the viscous calculation (blowing velocity)
        and use it as a boundary condition in the inviscid solver.
        """
        # Retreive and set blowing velocities.
        for iRegion in range(3):
            self.blwVel[iRegion] = np.asarray(self.vSolver.extractBlowingVel(0, iRegion))
            self.deltaStar[iRegion] = np.asarray(self.vSolver.extractDeltaStar(0, iRegion))
            self.xx[iRegion] = np.asarray(self.vSolver.extractxx(0, iRegion))
        blwVelAF = np.concatenate((self.blwVel[0], self.blwVel[1]))

        # Remove stagnation point doublon.
        deltaStarAF = np.concatenate((self.deltaStar[0], self.deltaStar[1][1:]))
        xxAF = np.concatenate((self.xx[0], self.xx[1][1:]))

        # Blowing velocity, displacement thickness and mesh (local coordinates) distributions in the wake.
        blwVelWK = self.blwVel[2]
        deltaStarWK = self.deltaStar[2]
        xxWK = self.xx[2]
        # Update data structures for information transfer.
        self.group[0].u = blwVelAF[self.group[0].connectListElems.argsort()]
        self.group[1].u = blwVelWK[self.group[1].connectListElems.argsort()]
        self.group[0].deltaStar = deltaStarAF[self.group[0].connectListNodes.argsort()]
        self.group[1].deltaStar = deltaStarWK[self.group[1].connectListNodes.argsort()]
        self.group[0].xx = xxAF[self.group[0].connectListNodes.argsort()]
        self.group[1].xx = xxWK[self.group[1].connectListNodes.argsort()]

        # Impose blowing velocities.
        for n in range(0, len(self.group)):
            for i in range(0, self.group[n].nE):
                self.group[n].boundary.setU(i, self.group[n].u[i])