#ifndef DBOUNDARYLAYER_H
#define DBOUNDARYLAYER_H

#include "vii.h"
#include "DClosures.h"
#include "DDiscretization.h"
#include "DEdge.h"

namespace vii
{

/**
 * @brief Boundary layer region upper/lower side or wake.
 * @author Paul Dechamps
 */
class VII_API BoundaryLayer
{

private:
    unsigned int const nVar = 5; ///< Number of variables of the partial differential problem.
    double nCrit = 9.0;          ///< Critical amplification factor.

public:
    Discretization *mesh; ///< 1D surface boundary layer mesh.
    Edge *blEdge;         ///< Quantites at the inviscid boundary (edge of the boundary layer).
    Closures *closSolver; ///< Closure relations class.
    std::string name;     ///< Name of the region.

    // Boundary layer variables.
    std::vector<double> u;         ///< Solution vector (theta, H, N, ue, Ct).
    std::vector<double> Ret;       ///< Reynolds number based on the momentum thickness (theta).
    std::vector<double> cd;        ///< Local dissipation coefficient.
    std::vector<double> cf;        ///< Local friction coefficient.
    std::vector<double> Hstar;     ///< Kinetic energy shape parameter (thetaStar/theta).
    std::vector<double> Hstar2;    ///< Density shape parameter (deltaStarStar/theta).
    std::vector<double> Hk;        ///< Kinematic shape parameter (int(1-u/ue d_eta)).
    std::vector<double> ctEq;      ///< Equilibrium shear stress coefficient (turbulent BL).
    std::vector<double> us;        ///< Equivalent normalized wall slip velocity.
    std::vector<double> delta;     ///< Boundary layer thickness.
    std::vector<double> deltaStar; ///< Dispacement thickness (int(1-rho*u/rhoe*ue d_eta)).

    // Transition related variables.
    std::vector<int> regime;             ///< Laminar (0) or turbulent (1) regime.
    std::vector<double> blowingVelocity; ///< Blowing velocity.
    size_t transMarker;                  ///< Marker id of the transition location.
    double xtrF;                         ///< Forced transition location in the global frame of reference.
    double xtr;                          ///< Transition location in the global frame of reference.
    double xoctr;                        ///< Transition location in % of the chord.

    BoundaryLayer(double _xtrF = -1.0);
    ~BoundaryLayer();

    // Boundary conditions.
    void setStagBC(double Re);
    void setWakeBC(double Re, std::vector<double> upperConditions, std::vector<double> lowerConditions);
    void setMesh(std::vector<double> x,
                 std::vector<double> y,
                 std::vector<double> z);

    // Getters.
    size_t getnVar() const { return nVar; };
    double getnCrit() const { return nCrit; };

    // Others
    void printSolution(size_t iPoint) const;
    void computeBlwVel();
};
} // namespace vii
#endif // DBOUNDARYLAYER_H
